@@ -1,4 +1,5 @@
 #!/usr/bin/env python
+# pylint: disable=E0611,E0213,E1102,C0103,E1101,W0613,R0913,R0904
 #
 # A library that provides a Python interface to the Telegram Bot API
 # Copyright (C) 2015 Leandro Toledo de Souza <leandrotoeldodesouza@gmail.com>
@@ -16,6 +17,7 @@
 # You should have received a copy of the GNU Lesser Public License
 # along with this program.  If not, see [http://www.gnu.org/licenses/].
 
+"""This module contains a object that represents a Telegram Bot"""
 
 import json
 try:
@@ -32,11 +34,27 @@ import logging
 from telegram import (User, Message, Update, UserProfilePhotos, TelegramError,
                       ReplyMarkup, InputFile, TelegramObject, NullHandler)
 
-h = NullHandler()
-logging.getLogger(__name__).addHandler(h)
+H = NullHandler()
+logging.getLogger(__name__).addHandler(H)
 
 
 class Bot(TelegramObject):
+    """This object represents a Telegram Bot.
+
+    Attributes:
+        id (int):
+        first_name (str):
+        last_name (str):
+        username (str):
+        name (str):
+
+    Args:
+        token (str):
+        **kwargs: Arbitrary keyword arguments.
+
+    Keyword Args:
+        base_url (Optional[str]):
+    """
 
     def __init__(self,
                  token,
@@ -48,35 +66,71 @@ class Bot(TelegramObject):
         else:
             self.base_url = base_url + self.token
 
-        self.log = logging.getLogger(__name__)
+        self.bot = None
 
-        try:
-            bot = self.getMe()
+        self.logger = logging.getLogger(__name__)
 
-            self.id = bot.id
-            self.first_name = bot.first_name
-            self.last_name = bot.last_name
-            self.username = bot.username
+    def info(func):
+        """
+        Returns:
+        """
+        @functools.wraps(func)
+        def decorator(self, *args, **kwargs):
+            """
+            decorator
+            """
+            if not self.bot:
+                self.getMe()
 
-            self.__auth = True
+            result = func(self, *args, **kwargs)
+            return result
+        return decorator
 
-            self.log.info('Starting bot %s' % self.name)
-        except TelegramError:
-            raise TelegramError({'message': 'Bad token'})
+    @property
+    @info
+    def id(self):
+        """int: """
+        return self.bot.id
+
+    @property
+    @info
+    def first_name(self):
+        """str: """
+        return self.bot.first_name
+
+    @property
+    @info
+    def last_name(self):
+        """str: """
+        return self.bot.last_name
+
+    @property
+    @info
+    def username(self):
+        """str: """
+        return self.bot.username
 
     @property
     def name(self):
+        """str: """
         return '@%s' % self.username
 
     def log(func):
+        """
+        Returns:
+          A telegram.Message instance representing the message posted.
+        """
         logger = logging.getLogger(func.__module__)
 
         @functools.wraps(func)
         def decorator(self, *args, **kwargs):
-            logger.debug('Entering: %s' % func.__name__)
+            """
+            decorator
+            """
+            logger.debug('Entering: %s', func.__name__)
             result = func(self, *args, **kwargs)
             logger.debug(result)
-            logger.debug('Exiting: %s' % func.__name__)
+            logger.debug('Exiting: %s', func.__name__)
             return result
         return decorator
 
@@ -87,8 +141,14 @@ class Bot(TelegramObject):
         """
         @functools.wraps(func)
         def decorator(self, *args, **kwargs):
+            """
+            decorator
+            """
             url, data = func(self, *args, **kwargs)
 
+            if not kwargs.get('chat_id'):
+                raise TelegramError('Invalid chat_id.')
+
             if kwargs.get('reply_to_message_id'):
                 reply_to_message_id = kwargs.get('reply_to_message_id')
                 data['reply_to_message_id'] = reply_to_message_id
@@ -100,8 +160,8 @@ class Bot(TelegramObject):
                 else:
                     data['reply_markup'] = reply_markup
 
-            json_data = self._requestUrl(url, 'POST', data=data)
-            data = self._parseAndCheckTelegram(json_data)
+            json_data = Bot._requestUrl(url, 'POST', data=data)
+            data = Bot._parseAndCheckTelegram(json_data)
 
             if data is True:
                 return data
@@ -109,22 +169,6 @@ class Bot(TelegramObject):
             return Message.de_json(data)
         return decorator
 
-    def require_authentication(func):
-        @functools.wraps(func)
-        def decorator(self, *args, **kwargs):
-            if not self.__auth:
-                raise TelegramError({'message': "API must be authenticated."})
-
-            return func(self, *args, **kwargs)
-        return decorator
-
-    @log
-    @require_authentication
-    def clearCredentials(self):
-        """Clear any credentials for this instance.
-        """
-        self.__auth = False
-
     @log
     def getMe(self):
         """A simple method for testing your bot's auth token.
@@ -138,17 +182,17 @@ class Bot(TelegramObject):
         json_data = self._requestUrl(url, 'GET')
         data = self._parseAndCheckTelegram(json_data)
 
-        return User.de_json(data)
+        self.bot = User.de_json(data)
+
+        return self.bot
 
     @log
     @message
-    @require_authentication
     def sendMessage(self,
                     chat_id,
                     text,
                     disable_web_page_preview=None,
-                    reply_to_message_id=None,
-                    reply_markup=None):
+                    **kwargs):
         """Use this method to send text messages.
 
         Args:
@@ -182,7 +226,6 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def forwardMessage(self,
                        chat_id,
                        from_chat_id,
@@ -216,13 +259,11 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def sendPhoto(self,
                   chat_id,
                   photo,
                   caption=None,
-                  reply_to_message_id=None,
-                  reply_markup=None):
+                  **kwargs):
         """Use this method to send photos.
 
         Args:
@@ -258,16 +299,23 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def sendAudio(self,
                   chat_id,
                   audio,
-                  reply_to_message_id=None,
-                  reply_markup=None):
+                  duration=None,
+                  performer=None,
+                  title=None,
+                  **kwargs):
         """Use this method to send audio files, if you want Telegram clients to
-        display the file as a playable voice message. For this to work, your
-        audio must be in an .ogg file encoded with OPUS (other formats may be
-        sent as telegram.Document).
+        display them in the music player. Your audio must be in an .mp3 format.
+        On success, the sent Message is returned. Bots can currently send audio
+        files of up to 50 MB in size, this limit may be changed in the future.
+
+        For backward compatibility, when both fields title and description are
+        empty and mime-type of the sent file is not "audio/mpeg", file is sent
+        as playable voice message. In this case, your audio must be in an .ogg
+        file encoded with OPUS. This will be removed in the future. You need to
+        use sendVoice method instead.
 
         Args:
           chat_id:
@@ -276,6 +324,12 @@ class Bot(TelegramObject):
             Audio file to send. You can either pass a file_id as String to
             resend an audio that is already on the Telegram servers, or upload
             a new audio file using multipart/form-data.
+          duration:
+            Duration of sent audio in seconds. [Optional]
+          performer:
+            Performer of sent audio. [Optional]
+          title:
+            Title of sent audio. [Optional]
           reply_to_message_id:
             If the message is a reply, ID of the original message. [Optional]
           reply_markup:
@@ -292,16 +346,21 @@ class Bot(TelegramObject):
         data = {'chat_id': chat_id,
                 'audio': audio}
 
+        if duration:
+            data['duration'] = duration
+        if performer:
+            data['performer'] = performer
+        if title:
+            data['title'] = title
+
         return url, data
 
     @log
     @message
-    @require_authentication
     def sendDocument(self,
                      chat_id,
                      document,
-                     reply_to_message_id=None,
-                     reply_markup=None):
+                     **kwargs):
         """Use this method to send general files.
 
         Args:
@@ -331,12 +390,10 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def sendSticker(self,
                     chat_id,
                     sticker,
-                    reply_to_message_id=None,
-                    reply_markup=None):
+                    **kwargs):
         """Use this method to send .webp stickers.
 
         Args:
@@ -366,14 +423,12 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def sendVideo(self,
                   chat_id,
                   video,
                   duration=None,
                   caption=None,
-                  reply_to_message_id=None,
-                  reply_markup=None):
+                  **kwargs):
         """Use this method to send video files, Telegram clients support mp4
         videos (other formats may be sent as telegram.Document).
 
@@ -414,13 +469,55 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
+    def sendVoice(self,
+                  chat_id,
+                  voice,
+                  duration=None,
+                  **kwargs):
+        """Use this method to send audio files, if you want Telegram clients to
+        display the file as a playable voice message. For this to work, your
+        audio must be in an .ogg file encoded with OPUS (other formats may be
+        sent as Audio or Document). On success, the sent Message is returned.
+        Bots can currently send audio files of up to 50 MB in size, this limit
+        may be changed in the future.
+
+        Args:
+          chat_id:
+            Unique identifier for the message recipient - User or GroupChat id.
+          voice:
+            Audio file to send. You can either pass a file_id as String to
+            resend an audio that is already on the Telegram servers, or upload
+            a new audio file using multipart/form-data.
+          duration:
+            Duration of sent audio in seconds. [Optional]
+          reply_to_message_id:
+            If the message is a reply, ID of the original message. [Optional]
+          reply_markup:
+            Additional interface options. A JSON-serialized object for a
+            custom reply keyboard, instructions to hide keyboard or to force a
+            reply from the user. [Optional]
+
+        Returns:
+          A telegram.Message instance representing the message posted.
+        """
+
+        url = '%s/sendVoice' % self.base_url
+
+        data = {'chat_id': chat_id,
+                'voice': voice}
+
+        if duration:
+            data['duration'] = duration
+
+        return url, data
+
+    @log
+    @message
     def sendLocation(self,
                      chat_id,
                      latitude,
                      longitude,
-                     reply_to_message_id=None,
-                     reply_markup=None):
+                     **kwargs):
         """Use this method to send point on the map.
 
         Args:
@@ -451,7 +548,6 @@ class Bot(TelegramObject):
 
     @log
     @message
-    @require_authentication
     def sendChatAction(self,
                        chat_id,
                        action):
@@ -468,8 +564,8 @@ class Bot(TelegramObject):
             is about to receive:
             - ChatAction.TYPING for text messages,
             - ChatAction.UPLOAD_PHOTO for photos,
-            - ChatAction.UPLOAD_VIDEO or upload_video for videos,
-            - ChatAction.UPLOAD_AUDIO or upload_audio for audio files,
+            - ChatAction.UPLOAD_VIDEO for videos,
+            - ChatAction.UPLOAD_AUDIO for audio files,
             - ChatAction.UPLOAD_DOCUMENT for general files,
             - ChatAction.FIND_LOCATION for location data.
         """
@@ -482,7 +578,6 @@ class Bot(TelegramObject):
         return url, data
 
     @log
-    @require_authentication
     def getUserProfilePhotos(self,
                              user_id,
                              offset=None,
@@ -518,7 +613,6 @@ class Bot(TelegramObject):
         return UserProfilePhotos.de_json(data)
 
     @log
-    @require_authentication
     def getUpdates(self,
                    offset=None,
                    limit=100,
@@ -557,17 +651,17 @@ class Bot(TelegramObject):
         data = self._parseAndCheckTelegram(json_data)
 
         if data:
-            self.log.info(
-                'Getting updates: %s' % [u['update_id'] for u in data])
+            self.logger.info(
+                'Getting updates: %s', [u['update_id'] for u in data])
         else:
-            self.log.info('No new updates found.')
+            self.logger.info('No new updates found.')
 
         return [Update.de_json(x) for x in data]
 
     @log
-    @require_authentication
     def setWebhook(self,
-                   webhook_url):
+                   webhook_url=None,
+                   certificate=None):
         """Use this method to specify a url and receive incoming updates via an
         outgoing webhook. Whenever there is an update for the bot, we will send
         an HTTPS POST request to the specified url, containing a
@@ -584,15 +678,19 @@ class Bot(TelegramObject):
         """
         url = '%s/setWebhook' % self.base_url
 
-        data = {'url': webhook_url}
+        data = {}
+        if webhook_url:
+            data['url'] = webhook_url
+        if certificate:
+            data['certificate'] = certificate
 
         json_data = self._requestUrl(url, 'POST', data=data)
         data = self._parseAndCheckTelegram(json_data)
 
-        return True
+        return data
 
-    def _requestUrl(self,
-                    url,
+    @staticmethod
+    def _requestUrl(url,
                     method,
                     data=None):
         """Request an URL.
@@ -629,12 +727,12 @@ class Bot(TelegramObject):
                         url,
                         urlencode(data).encode()
                     ).read()
-            except IOError as e:
-                raise TelegramError(str(e))
             except HTTPError as e:
                 raise TelegramError(str(e))
             except URLError as e:
                 raise TelegramError(str(e))
+            except IOError as e:
+                raise TelegramError(str(e))
 
         if method == 'GET':
             try:
@@ -642,8 +740,8 @@ class Bot(TelegramObject):
             except URLError as e:
                 raise TelegramError(str(e))
 
-    def _parseAndCheckTelegram(self,
-                               json_data):
+    @staticmethod
+    def _parseAndCheckTelegram(json_data):
         """Try and parse the JSON returned from Telegram and return an empty
         dictionary if there is any error.
 
@@ -666,7 +764,15 @@ class Bot(TelegramObject):
 
         return data['result']
 
+    @staticmethod
+    def de_json(data):
+        pass
+
     def to_dict(self):
+        """
+        Returns:
+            dict:
+        """
         data = {'id': self.id,
                 'username': self.username,
                 'first_name': self.username}
